#------------------------------------------------------------------
# $Copyright: Copyright (c) 2020 Veritas Technologies LLC.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# VERITAS TECHNOLOGIES LLC.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF VERITAS TECHNOLOGIES LLC.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/SAPMaxDB/SAPMaxDB.pm,v 1.15 2014/06/23 07:53:35 avaidya Exp $
#
# $Author: avaidya $
# $Date: 2014/06/23 07:53:35 $
# $Revision: 1.15 $
#
# Provides:                            SAPMaxDB
# Author:                              vchappav
# crAgent Build:                       1.62 2007/07/03 09:31:12
# crAgent Run Date:                    13 Jul, 2007
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package SAPMaxDB;

use IPC::SysV qw ( IPC_STAT IPC_RMID );
my $sCmdIpcs   = "/usr/bin/ipcs";

my ( $CLUS_HOME, $sPrefix ) = ();

#---------------------
# Setup VCS/VCSOne home..
#---------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.0/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vcs(|one))$/;

if ( $sPrefix =~ /vcsone/ ) {
  $sPrefix = "ha";
}

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = "$CLUS_HOME/bin/hares";
my $sCmdHatype = "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";

my @liRunningProcs = (); # Global..

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use SystemD;        # Support generalized SystemD Functions
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::stat;

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: SAPMaxDB.pm,v 1.15 2014/06/23 07:53:35 avaidya Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define SAPMaxDB subroutines..
#--------------------------------

=pod

 -------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String EntryPointName,
                              IN String EnvFile,
                              IN String DBHome,
                              IN String DBUser,
                              IN String DBName,
                              IN String DBHost,
                              IN String DBMUserId,
                              IN String DBMPassword,
                              IN Integer SecondLevelMonitor,
                              IN String MonitorProgram );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry pt name
 $localEnvFile                  Environments file for 'DBUser'. 
 $localDBHome                   Directory for MaxDB database software.
 $localDBUser                   Operating system user for MaxDB server.
 $localDBName                   Name of the MaxDB instance.
 $localDBHost                   Virtual host name of the MaxDB database server.
 $localDBMUserId                Database user ID used to connect to MaxDB database server.
 $localDBMPassword              Password for the user 'DBMUserId'.
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring.
 $localMonitorProgram           Full path to external monitor program.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localEnvFile,
       $localDBHome,
       $localDBUser,
       $localDBName,
       $localDBHost,
       $localDBMUserId,
       $localDBMPassword,
       $localSecondLevelMonitor,
       $localMonitorProgram ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, "Subroutine <%s> called with:
  EntryPointName           [%s]
  EnvFile                  [%s]
  DBHome                   [%s]
  DBUser                   [%s]
  DBName                   [%s]
  DBHost                   [%s]
  DBMUserId                [%s]
  DBMPassword              [%s]
  SecondLevelMonitor       [%s]
  MonitorProgram           [%s]",

  $sMyName,    $sEntryPointName,    $localEnvFile,    $localDBHome,
  $localDBUser,    $localDBName,   $localDBHost,   $localDBMUserId,
  $localDBMPassword,                      $localSecondLevelMonitor,
  $localMonitorProgram );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName EnvFile  DBHome  DBUser  DBName  DBHost  DBMUserId
       DBMPassword SecondLevelMonitor MonitorProgram );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( EnvFile DBHome DBUser DBName DBHost DBMUserId DBMPassword );
  my @lsOfflineArgs = qw ( EnvFile DBHome DBUser DBName DBMUserId DBMPassword );
  my @lsMonitorArgs = qw ( DBName );
  my @lsCleanArgs   = qw ( DBName );
  #my @lsMonitorArgs = qw ( EnvFile DBHome DBUser DBName DBMUserId DBMPassword );
  #my @lsCleanArgs   = qw ( EnvFile DBHome DBUser DBName DBMUserId DBMPassword );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{$hlsNeeded { $sEntryPointName }};

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  #----------------------------------
  # Validation for all entry points..
  #----------------------------------
  #--------------------------
  # Validate parameter DBName..
  #--------------------------
  if ( $localDBName !~ /^([A-Z][A-Z0-9_]{0,7})$/ ) {
    &logError ( 20003, "<%s> DBName [%s] contains invalid characters",
    $sMyName, $localDBName );
    $iReturnCode = 0; # Failure..
  }

  #---------------------------------------------------
  # Validation for online offline entry point(s)..
  #---------------------------------------------------
  if ( $sEntryPointName =~ /^(online|offline)$/o ) {
    #--------------------------
    # Validate parameter DBUser..
    #--------------------------
    if ( ! defined getpwnam $localDBUser ) {
      &logError ( 20004, "<%s> DBUser [%s] is not a valid UNIX login",
      $sMyName, $localDBUser );
      $iReturnCode = 0; # Failure..
    }

    #---------------------------
    # Validate parameter EnvFile..
    #---------------------------
    if ( ! -f  $localEnvFile ) {
      &logError ( 20005, "<%s> EnvFile [%s] is not a plain file",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }

    #--------------------------
    # Validate parameter DBHome..
    #--------------------------
    if ( ! -d  $localDBHome ) {
      &logError ( 20006, "<%s> DBHome [%s] directory does not exist",
      $sMyName, $localDBHome );
      $iReturnCode = 0; # Failure..
    }
  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #----------------------------
    # Validate parameter DBHost..
    #----------------------------
    if ( ! inet_aton $localDBHost ) {
      &logError ( 20007, "<%s> DBHost [%s] is an invalid Hostname/IPv4 address",
      $sMyName, $localDBHost );
      $iReturnCode = 0; # Failure..
    }

    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20008, "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }

  }

  $iReturnCode ||
  &logWarn ( 20009,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}



=pod

 -------------------------------------------------------------------------
 Function : bool Start ( IN String DBHome,
                         IN String DBHost,
                         IN String DBMPassword,
                         IN String DBMUserId,
                         IN String DBName,
                         IN String DBUser,
                         IN String EnvFile,
                         IN Integer SecondLevelMonitor );

 Description : Start logic needed for online entry pt

 Parameters :

 $sDBHome                   Directory for MaxDB database software.
 $sDBHost                   Virtual host name of the MaxDB database server.
 $sDBMPassword              Password for the user 'DBMUserId'.
 $sDBMUserId                Database user ID used to connect to MaxDB database server.
 $sDBName                   Name of the MaxDB instance.
 $sDBUser                   Operating system user for MaxDB server.
 $sEnvFile                  Environments file for 'DBUser'. 
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBHome,
       $sDBHost,
       $sDBMPassword,
       $sDBMUserId,
       $sDBName,
       $sDBUser,
       $sEnvFile,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20010, "Subroutine <%s> called with:
  DBHome                   [%s]
  DBHost                   [%s]
  DBMPassword              [%s]
  DBMUserId                [%s]
  DBName                   [%s]
  DBUser                   [%s]
  EnvFile                  [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,   $sDBHome,   $sDBHost,   $sDBMPassword,   $sDBMUserId,
  $sDBName, $sDBUser, $sEnvFile, $iSecondLevelMonitor );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, $iTimeout,
       @lsCmdout, @liPids ) = ();

  #-----------------------------------------
  # Whack any preexisting SAPMaxDB processes..
  #-----------------------------------------
  &logTrace ( 20011,
  "<%s> Killing SAPMaxDB processes for this instance", $sMyName );

  if ( SAPMaxDB::KillSAPMaxDBProcesses ( $sDBName,
                                         $sDBUser ) ) {
    &logTrace ( 20012,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20013,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  #------------------------------------------------------
  # Remove any preexisting SAPMaxDB IPC resources...
  #------------------------------------------------------
  &logTrace ( 20014,
  "<%s> Cleaning SAPMaxDB IPC resources for instance [%s]", $sMyName, $sDBName );

  if(SAPMaxDB::RemoveIPCResourcesForSAPMaxDB ( "sdb",
                                               $sDBHome,
                                               $sDBName ) ) {
    &logTrace(20015,
    "<%s> Removed all IPC resources for SAP MaxDB Instance [%s] successfully",
    $sMyName, $sDBName );
  }
  else {
    &logError(20016,
    "<%s> Removed all IPC resources for SAP MaxDB Instance [%s] successfully",
    $sMyName, $sDBName );
    $iReturnCode = 0; # Failure ..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + SAPMaxDB::GetEngineTimeout ( 1 );

  #----------------------------------------------------
  # Start x_server for SAP MaxDB Server..
  #----------------------------------------------------
  $sCommand = "$sDBHome/x_server";
  $sCommandArgs = "start";

  &logTrace ( 20017,
  "<%s> Starting x_server with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20018,
    "<%s> Insufficient time to start x_server. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = 0; # Failure..
  }

  elsif( $::gbUseSystemD && SystemD::DetectSystemd() ) {
    if( $iReturnCode = SystemD::RunSystemdWithEnv( $sDBUser , $sCommand, $sCommandArgs, $sEnvFile, $iTimeout,"x_server", undef,undef) ){
      &logTrace (20234,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }
  
  elsif ( $iReturnCode =
    SAPMaxDB::ExecLocalCmd ( $sDBUser,
                             $sEnvFile,
                             $sCommand,
                             $sCommandArgs,
                             \$iExitCode,
                             $iTimeout,
                             \@lsCmdout ) ) {
    &logTrace ( 20019,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    my $sSrchStr = qr|\bXSERVER\b.*\bstarted\b|io; # Search pattern..
    my $sSrchStr1 = qr|\bFound\s+running\s+XServer\b|o; # Search pattern..

    &logTrace ( 20228, "<%s> Looking with filter [%s] or [%s] from output",
    $sMyName, $sSrchStr, $sSrchStr1 );

    #--------------------------------------------
    # Attempt to locate search string in output..
    #--------------------------------------------
    unless ( grep { /$sSrchStr/ || /$sSrchStr1/ } @lsCmdout ) {
      &logError ( 20229,
      "<%s> Failed to locate search pattern [%s] and [%s] in Cmd output [%s]",
      $sMyName, $sSrchStr, $sSrchStr1, join ( chr 012, @lsCmdout ) );

      $iReturnCode = 0; # Failure..
    }
  }
  else {
    &logError ( 20022,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );
    $iReturnCode = 0; # Failure..
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;

  #------------------------------------------
  # Define start command and args variables..
  #------------------------------------------
  $sCommand = "$sDBHome/dbmcli";
  $sCommandArgs = "-d $sDBName -n $sDBHost -u $sDBMUserId,$sDBMPassword db_online";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20023,
  "<%s> Starting SAPMaxDB instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20024,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }

 elsif( $::gbUseSystemD && SystemD::DetectSystemd() ) {
  $sCommandArgs = "-d $sDBName -n $sDBHost -u $sDBMUserId,XXXXX db_online";  #Replacing sDBMPassword with XXXXX
  if( $iReturnCode = SystemD::RunSystemdWithEnv( $sDBUser , $sCommand, $sCommandArgs, $sEnvFile, $iTimeout,"dbmcli", undef, $sDBMPassword) ){
      &logTrace (20235,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }

  elsif ( $iReturnCode =
    SAPMaxDB::ExecLocalCmd ( $sDBUser, 
                             $sEnvFile, 
                             $sCommand, 
                             $sCommandArgs, 
                             \$iExitCode, 
                             $iTimeout, 
                             \@lsCmdout, 
                             $sDBMPassword ) ) {
    &logTrace ( 20025,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logError ( 20026,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;


  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20027,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         SAPMaxDB::GetProcState ( \@liPids,
                                  $sDBName,
                                  $sDBUser,
                                  1 ) ) {
      #---------------------
      # Start-up confirmed..
      #---------------------
      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20028,
      "<%s> Could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0;  # Failure..
    }
  }

  if ( $iReturnCode && $iSecondLevelMonitor ) {  # If enabled..
    while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
      sleep $iSleepTime; # Sec..

      &logTrace ( 20029,
      "<%s> Confirming start-up with [%s] sec in hand",
      $sMyName, $iTimeout );

      #---------
      # Do SLM..
      #---------
      if ( VCSDef_MonitorExit::ONLINE ==              # Online after..
           SAPMaxDB::SecondLevelMonitor ( $sDBHome,   # Running SLM..
                                          $sDBMPassword,
                                          $sDBMUserId,
                                          $sDBName,
                                          $sDBUser,
                                          $sEnvFile,
                                          $iTillWhen ) ) { 
        #---------------------
        # Start-up confirmed..
        #---------------------
        &logInfo ( 20030,
        "<%s> Confirmed start-up of this instance of SAPMaxDB server",
        $sMyName );

        $iReturnCode = 1; # Success..
        last;
      }
      else {
        &logTrace ( 20031,
        "<%s> Could not confirm start-up. Sleeping for [%s] sec",
        $sMyName, $iSleepTime );

        $iReturnCode = 0;  # Failure..
      }
    }
  }
  elsif ( $iReturnCode && !$iSecondLevelMonitor ) {   # If only FLM succeeds .. 
    &logInfo ( 20032,
    "<%s> Confirmed start-up of this instance of SAPMaxDB server",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Stop ( IN String DBHome,
                        IN String DBMPassword,
                        IN String DBMUserId,
                        IN String DBName,
                        IN String DBUser,
                        IN String EnvFile );

 Description : Stop logic needed for offline entry pt

 Parameters :

 $sDBHome                   Directory for MaxDB database software.
 $sDBMPassword              Password for the user 'DBMUserId'.
 $sDBMUserId                Database user ID used to connect to MaxDB database server.
 $sDBName                   Name of the MaxDB instance.
 $sDBUser                   Operating system user for MaxDB server.
 $sEnvFile                  Environments file for 'DBUser'. 
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBHome,
       $sDBMPassword,
       $sDBMUserId,
       $sDBName,
       $sDBUser,
       $sEnvFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20033, "Subroutine <%s> called with:
  DBHome                   [%s]
  DBMPassword              [%s]
  DBMUserId                [%s]
  DBName                   [%s]
  DBUser                   [%s]
  EnvFile                  [%s]",

  $sMyName,   $sDBHome,   $sDBMPassword,   $sDBMUserId,   $sDBName,
  $sDBUser, $sEnvFile );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPMaxDB::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, $iSleepTime,
       @lsCmdout, @liPids ) = ();
  
  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  $sCommand = "$sDBHome/dbmcli";
  $sCommandArgs = "-d $sDBName -u $sDBMUserId,$sDBMPassword db_offline";
  ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 

  &logTrace ( 20034,
  "<%s> Stopping SAPMaxDB with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20035,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          SAPMaxDB::ExecLocalCmd ( $sDBUser, 
                                   $sEnvFile, 
                                   $sCommand, 
                                   $sCommandArgs, 
                                   \$iExitCode, 
                                   $iTimeout, 
                                   \@lsCmdout, 
                                   $sDBMPassword ) ) {
    &logTrace ( 20036,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logError ( 20037,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );
  }

  if ($::gbStopGlobalListeners) {
    #----------------------------------------------------
    # Stop x_server for SAP MaxDB Server..
    #----------------------------------------------------
    $sCommand = "$sDBHome/sdbgloballistener";
    $sCommandArgs = "stop -all";
  
    &logTrace ( 20236,
    "<%s> Stopping the global listeners with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------
    # Request shutdown..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20237,
      "<%s> Insufficient time to stop all the sdbgloballisteners. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
      SAPMaxDB::ExecLocalCmd ( $sDBUser,
                               $sEnvFile,
                               $sCommand,
                               $sCommandArgs,
                               \$iExitCode,
                               $iTimeout,
                               \@lsCmdout ) ) {
      &logTrace ( 20238,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
      if ( 0 == $iExitCode )
      {
        &logTrace ( 20239,
        "<%s> [%s] with args [%s] executed successfully with ExitCode [%s] and output [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
      }
      else
      {
        &logError ( 20240,
        "<%s> [%s] with args [%s] failed with ExitCode [%s] and output [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
      }
    }
    else {
      &logError ( 20241,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
      Cmd output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode,
      join ( chr 012, @lsCmdout ) );
    }
  }


  #--------------------
  # Delay entry point..
  #--------------------
  if ( $iReturnCode ) { # Success..
    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20038,
    "<%s> Delaying exit to confirm shutdown...", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      SAPMaxDB::GetProcState ( \@liPids,
                               $sDBName,
                               $sDBUser,
                               0 ); # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20039,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout );
      }
      else {
        &logInfo ( 20040,
        "<%s> Confirmed shutdown of SAPMaxDB instance",
        $sMyName );
        last; # Get out..
      }
    }
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;


  #-----------------------------------------
  # Whack any persisting SAPMaxDB processes..
  #-----------------------------------------
  &logTrace ( 20041,
  "<%s> Killing SAPMaxDB processes for this instance", $sMyName );

  if ( SAPMaxDB::KillSAPMaxDBProcesses ( $sDBName,
                                         $sDBUser ) ) {
    &logTrace ( 20042,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20043,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  #------------------------------------------------------
  # Remove any left over SAPMaxDB IPC resources...
  #------------------------------------------------------
  &logTrace ( 20044,
  "<%s> Cleaning SAPMaxDB IPC resources for instance [%s]", $sMyName, $sDBName );

  if(SAPMaxDB::RemoveIPCResourcesForSAPMaxDB ( "sdb",
                                               $sDBHome,
                                               $sDBName ) ) {
    &logTrace(20045,
              "<%s> Removed all IPC resources for SAP MaxDB Instance [%s] successfully",
              $sMyName, $sDBName );
  }
  else {
    &logError(20046,
              "<%s> Removed all IPC resources for SAP MaxDB Instance [%s] successfully",
              $sMyName, $sDBName );
    $iReturnCode = 0; # Failure ..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool Kill ( IN Integer CleanReason,
                        IN String DBHome,
                        IN String DBMPassword,
                        IN String DBMUserId,
                        IN String DBName,
                        IN String DBUser,
                        IN String EnvFile );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $iCleanReason              Why clean was called
 $sDBHome                   Directory for MaxDB database software.
 $sDBMPassword              Password for the user 'DBMUserId'.
 $sDBMUserId                Database user ID used to connect to MaxDB database server.
 $sDBName                   Name of the MaxDB instance.
 $sDBUser                   Operating system user for MaxDB server.
 $sEnvFile                  Environments file for 'DBUser'. 
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iCleanReason,
       $sDBHome,
       $sDBMPassword,
       $sDBMUserId,
       $sDBName,
       $sDBUser,
       $sEnvFile ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out      Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20047, "Subroutine <%s> called with:
  CleanReason              [%s]
  DBHome                   [%s]
  DBMPassword              [%s]
  DBMUserId                [%s]
  DBName                   [%s]
  DBUser                   [%s]
  EnvFile                  [%s]",

  $sMyName, $sCleanReason,  $sDBHome,  $sDBMPassword,  $sDBMUserId,
  $sDBName, $sDBUser, $sEnvFile );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPMaxDB::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  $sCommand = "$sDBHome/dbmcli";
  $sCommandArgs = "-d $sDBName -u $sDBMUserId,$sDBMPassword db_offline";

  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20048,
    "<%s> Attempting to stop SAPMaxDB with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20049,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode = SAPMaxDB::ExecLocalCmd ( $sDBUser, 
                                                    $sEnvFile, 
                                                    $sCommand, 
                                                    $sCommandArgs, 
                                                    \$iExitCode, 
                                                    $iTimeout, 
                                                    \@lsCmdout, 
                                                    $sDBMPassword ) ) {
      &logTrace ( 20050, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      &logError ( 20051,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
      Cmd output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode,
      join ( chr 012, @lsCmdout ) );
    }

    if ($::gbStopGlobalListeners) {
      #----------------------------------------------------
      # Stop x_server for SAP MaxDB Server..
      #----------------------------------------------------
      $sCommand = "$sDBHome/sdbgloballistener";
      $sCommandArgs = "stop -all";
    
      &logTrace ( 20242,
      "<%s> Stopping the sdbgloballisteners with Cmd [%s] Args [%s]",
      $sMyName, $sCommand, $sCommandArgs );
    
      #------------------
      # Request startup..
      #------------------
      if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
        &logError ( 20243,
        "<%s> Insufficient time to stop sdbgloballisteners. Short by [%s] sec",
        $sMyName, abs ( $iTimeout ) );
        $iReturnCode = 0; # Failure..
      }
      elsif ( $iReturnCode =
        SAPMaxDB::ExecLocalCmd ( $sDBUser,
                                 $sEnvFile,
                                 $sCommand,
                                 $sCommandArgs,
                                 \$iExitCode,
                                 $iTimeout,
                                 \@lsCmdout ) ) {
        &logTrace ( 20244,
        "<%s> Cmd [%s] with args [%s] executed successfully",
        $sMyName, $sCommand, $sCommandArgs );
        if ( 0 == $iExitCode )
        {
          &logTrace ( 20245,
          "<%s> [%s] with args [%s] executed successfully with ExitCode [%s] and output [%s]",
          $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
        }
        else
        {
          &logError ( 20246,
          "<%s> [%s] with args [%s] failed with ExitCode [%s] and output [%s]",
          $sMyName, $sCommand, $sCommandArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
        }
      }
      else {
        &logError ( 20247,
        "<%s> Cmd [%s] with args [%s] failed with exit code [%s].
        Cmd output [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode,
        join ( chr 012, @lsCmdout ) );
      }
    }
  }

  #-----------------------------------------
  # Whack any persisting SAPMaxDB processes..
  #-----------------------------------------
  &logTrace ( 20052,
  "<%s> Killing SAPMaxDB processes for this instance", $sMyName );

  if ( SAPMaxDB::KillSAPMaxDBProcesses ( $sDBName,
                                         $sDBUser ) ) {
    &logTrace ( 20053,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20054,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  #-------------------------------------------
  # Define clear command and args variables..
  #-------------------------------------------
  $sCommand = "$sDBHome/dbmcli";
  $sCommandArgs = "-d $sDBName -u $sDBMUserId,$sDBMPassword db_clear";

  #-------------------
  # Try to be gentle..
  #-------------------
  &logInfo ( 20055,
  "<%s> Attempting to clear SAPMaxDB with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #-----------------
  # Request clear ..
  #-----------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20056,
    "<%s> Insufficient time to continue clear. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    #$iReturnCode = 0; # Failure..
  }
  elsif ( SAPMaxDB::ExecLocalCmd ( $sDBUser,
                                   $sEnvFile,
                                   $sCommand,
                                   $sCommandArgs,
                                   \$iExitCode,
                                   $iTimeout,
                                   \@lsCmdout,
                                   $sDBMPassword ) ) {
    &logTrace ( 20057,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #------------------------------------------------------
  # Remove any left over SAPMaxDB IPC resources...
  #------------------------------------------------------
  &logTrace ( 20058,
  "<%s> Cleaning SAPMaxDB IPC resources for instance [%s]", $sMyName, $sDBName );

  if( $iReturnCode = SAPMaxDB::RemoveIPCResourcesForSAPMaxDB ( "sdb",
                                                               $sDBHome,
                                                               $sDBName ) ) {
    &logTrace(20059,
    "<%s> Removed all IPC resources for SAP MaxDB Instance [%s] successfully",
    $sMyName, $sDBName );
  }
  else {
    &logError(20060,
    "<%s> Failed to Remove all IPC resources for SAP MaxDB Instance [%s]",
    $sMyName, $sDBName );

    $iReturnCode = 0; # Failure ..
  }

  return $iReturnCode;
}



=pod

 -------------------------------------------------------------------------
 Function : int GetResourceState ( IN String DBName,
                                   IN String DBUser );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sDBName                   Name of the MaxDB instance.
 $sDBUser                   Operating system user for MaxDB server.
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBName,
       $sDBUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20061, "Subroutine <%s> called:
  DBName                   [%s]
  DBUser                   [%s]",

  $sMyName, $sDBName, $sDBUser );

  my @liDummyPids = ();
  my $iReturnCode = SAPMaxDB::GetProcState ( \@liDummyPids,
                                             $sDBName,
                                             $sDBUser,
                                             1 ); # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20062,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20063,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20064,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}


=pod

 -------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN String DBName,
                               IN String DBUser,
                               IN Integer UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sDBName                   Name of the MaxDB instance.
 $sDBUser                   Operating system user for MaxDB server.
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50   VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sDBName,
       $sDBUser,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20065, "Subroutine <%s> called with:
  OnlinePids               [%s]
  DBName                   [%s]
  DBUser                   [%s]
  UseFFLM                  [%s]",

  $sMyName, $rliOnlinePids, $sDBName, $sDBUser, $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  $sDBUser = "sdb"; # Change the user to MaxDB process owner..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();
  my %hiDBPid = ();
  my $iPPid = ();  

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------
  if ( ( 0 == ( $iUid = getpwnam $sDBUser ) ) && # User is the super-user..
       ( 2 == $iUseFFLM ) ) {                  # And called to kill-by-UID..
    &logTrace ( 20066, "<%s> Unsafe to kill processes by User [%s] UID [%s].
    Using filter-based approach.", $sMyName, $sDBUser, $iUid ),
    $iUseFFLM = 1;
  }

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( SAPMaxDB::GetSrvrDetails ( $sDBName, 
                                    \%hsSrvrDetails ) ) {
      &logTrace ( 20067,
      "<%s> Got details for monitoring SAPMaxDB instance", $sMyName );
  
      %hsCopy = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20068,
      "<%s> Failed to get details for monitoring SAPMaxDB instance",
      $sMyName );
      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp .= "\nSAPMaxDB process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = SAPMaxDB::FFLMStatus ( $sPidFile,
                                           $sProc,
                                           $iUid,
                                           \$iPid,
                                           \$iPPid );
          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
          $hiProcPids { $sKey } = $iPid;   # Store..

          $hiDBPid { $iPid } = $iPPid;    # by vchappav ...
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20069,
      "<%s> SAPMaxDB instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20070,
    "<%s> Called to get processes for User [%s] UID [%s]",
    $sMyName, $sDBUser, $iUid );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20071,
    "<%s> Running regular first level monitor", $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );
    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20072,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20073,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      &logTrace ( 20074, "<%s> Searching PID [%s] UID [%s] CMD [%s]",
      $sMyName, $$rhlProc { PID }, $$rhlProc { UID },
      $$rhlProc { COMMAND } );

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      if ( $$rhlProc { UID } == $iUid ) { # Process owner..
        #-----------------------------------------------
        # Look for processes that were offline in FFLM..
        #-----------------------------------------------
        if ( $iUseFFLM != 2 ) { # Request to filter..
          foreach $sKey ( keys %hsCopy ) {
            @lsUpdates = (); # Reset..
            @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
            #--------------------------------------------------
            # Should have details per three elements of array..
            # <PIDFILE> <PROCESS> <STATE>
            #--------------------------------------------------
            while ( ( $sPidFile, $sProc, $iFound ) =
                      splice ( @lsValues, 0, 3 ) ) {
              if ( ( VCSDef_MonitorExit::OFFLINE == $iFound ) && # Earlier marked offline..
                   ( $$rhlProc { COMMAND } =~ /$sProc\b/ ) &&  # Now found..  by vchappav
                   ( ! defined $hiDBPid { $iPid } )      &&     # Not already matched..
		   (($$rhlProc { PPID } == 1 ) || ($sProc !~ /vserver -sdbstarter/))) {

                $hiDBPid { $iPid } = $$rhlProc { PPID };         # Store PPID of the process..

                &logTrace ( 20075,
                "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND }, $sProc );

                $hiProcPids { $sKey } = $iPid;        # Store..
                unshift ( @$rliOnlinePids, $iPid );   # Store..
                $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..

                &logTrace ( 20076,
                "<%s> Got [%s] processes into hiDBPid", $sMyName, scalar(keys(%hiDBPid)) );

                #----------------
                # Update status..
                #----------------
                push ( @lsUpdates, $sPidFile, $sProc, $iFound );
                @{ $hsCopy { $sKey } } = @lsUpdates; # Update..

                last;
              }
            }
          }
        }
        else { # Request for kill..
          &logTrace ( 20077,
          "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
          $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
          push ( @$rliOnlinePids, $iPid ); # Store..

          $hiDBPid { $iPid } = $$rhlProc { PPID };         # Store PPID of the process...
        } 
      }
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20078,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( SAPMaxDB::CreatePidFile ( $sPidFile,
                                       $hiProcPids { $sKey } ) ) {
          &logTrace ( 20079,
          "<%s> Successfully created PID file [%s]", $sMyName, $sPidFile );
        }
        else {
          &logError ( 20080,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound && ( 2 <= scalar (keys (%hiDBPid)) ) ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20081,
    "<%s> Found [%s] PIDs [%s] for this instance of SAPMaxDB.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20082,
    "<%s> SAPMaxDB process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  # Sort the PIDs of the MaxDB processes in the descending order of their parent PIDs.
  if( (scalar @$rliOnlinePids > 0) && (scalar @$rliOnlinePids == scalar (keys (%hiDBPid)) ) ) {
    &logTrace ( 20083,
    "<%s> Sorting the MaxDB process IDs in the descending order of their parent IDs.",
    $sMyName );

    @$rliOnlinePids = sort {$hiDBPid{$b} <=> $hiDBPid{$a} } keys %hiDBPid ;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int Monitor ( IN String DBHome,
                          IN String DBMPassword,
                          IN String DBMUserId,
                          IN String DBName,
                          IN String DBUser,
                          IN String EnvFile,
                          IN String MonitorProgram,
                          IN Integer SecondLevelMonitor );

 Description: Performs the in-depth monitoring

 Parameters :

 $sDBHome                   Directory for MaxDB database software.
 $sDBMPassword              Password for the user 'DBMUserId'.
 $sDBMUserId                Database user ID used to connect to MaxDB database server.
 $sDBName                   Name of the MaxDB instance.
 $sDBUser                   Operating system user for MaxDB server.
 $sEnvFile                  Environments file for 'DBUser'. 
 $sMonitorProgram           Full path to external monitor program.
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring.
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
 -------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBHome,
       $sDBMPassword,
       $sDBMUserId,
       $sDBName,
       $sDBUser,
       $sEnvFile,
       $sMonitorProgram,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20084, "Subroutine <%s> called with:
  DBHome                   [%s]
  DBMPassword              [%s]
  DBMUserId                [%s]
  DBName                   [%s]
  DBUser                   [%s]
  EnvFile                  [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]",

  $sMyName,   $sDBHome,   $sDBMPassword,   $sDBMUserId,   $sDBName,
  $sDBUser, $sEnvFile, $sMonitorProgram, $iSecondLevelMonitor );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPMaxDB::DelayedAttributeCheck ( $sDBHome,
                                           $sDBMPassword,
                                           $sDBMUserId,
                                           $sDBName,
                                           $sDBUser,
                                           $sEnvFile ) ) ) {
      &logTrace ( 20085,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20086,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20087,
    "<%s> Proceeding with second level monitoring", $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + SAPMaxDB::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         SAPMaxDB::SecondLevelMonitor ( $sDBHome,
                                        $sDBMPassword,
                                        $sDBMUserId,
                                        $sDBName,
                                        $sDBUser,
                                        $sEnvFile,
                                        $iTillWhen ) ) ) {

      &logTrace ( 20088,
      "<%s> SAPMaxDB instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20089,
      "<%s> SAPMaxDB instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + SAPMaxDB::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         SAPMaxDB::ExternalMonitor ( $sDBUser,
                                     $sEnvFile,
                                     $sExtCmd,
                                     $sExtCmdArgs,
                                     $iTillWhen ) ) ) {
      &logTrace ( 20090,
      "<%s> SAPMaxDB instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20091,
      "<%s> SAPMaxDB instance NOT online after executing external monitor
      [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String DBHome,
                                        IN String DBMPassword,
                                        IN String DBMUserId,
                                        IN String DBName,
                                        IN String DBUser,
                                        IN String EnvFile );

 Description : Just In Time Validation

 Parameters :

 $sDBHome                   Directory for MaxDB database software.
 $sDBMPassword              Password for the user 'DBMUserId'.
 $sDBMUserId                Database user ID used to connect to MaxDB database server.
 $sDBName                   Name of the MaxDB instance.
 $sDBUser                   Operating system user for MaxDB server.
 $sEnvFile                  Environments file for 'DBUser'. 
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 -------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBHome,
       $sDBMPassword,
       $sDBMUserId,
       $sDBName,
       $sDBUser,
       $sEnvFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20092, "Subroutine <%s> called with:
  DBHome                   [%s]
  DBMPassword              [%s]
  DBMUserId                [%s]
  DBName                   [%s]
  DBUser                   [%s]
  EnvFile                  [%s]",

  $sMyName,   $sDBHome,   $sDBMPassword,   $sDBMUserId,   $sDBName,
  $sDBUser, $sEnvFile );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------
  # Validate parameter DBMUserId..
  #--------------------------------
  if ( $sDBMUserId !~ /\S/o ) {
    &logError ( 20093, "<%s> DBMUserId [%s] is null",
    $sMyName, $sDBMUserId );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #----------------------------------
  # Validate parameter DBMPassword..
  #----------------------------------
  if ( $sDBMPassword !~ /\S/o ) {
    &logError ( 20094, "<%s> DBMPassword [%s] is null",
    $sMyName, $sDBMPassword );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------------------
  # Validate parameter DBUser..
  #-----------------------------
  if ( $sDBUser !~ /\S/o ) {
    &logError ( 20095, "<%s> DBUser [%s] is null",
    $sMyName, $sDBUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( ! defined getpwnam $sDBUser ) {
    &logError ( 20096, "<%s> [%s] is not a valid UNIX login",
    $sMyName, $sDBUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( $sEnvFile !~ /\S/o ) {
    &logError ( 20097, "<%s> EnvFile [%s] is null",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( ! -f  $sEnvFile ) {
    &logError ( 20098, "<%s> [%s] is not a plain file",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------------------
  # Validate parameter DBHome..
  #-----------------------------
  if ( $sDBHome !~ /\S/o ) {
    &logError ( 20099, "<%s> DBHome [%s] directory null",
    $sMyName, $sDBHome );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( ! -d  $sDBHome ) {
    &logError ( 20100, "<%s> [%s] directory does not exist",
    $sMyName, $sDBHome );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String DBHome,
                                     IN String DBMPassword,
                                     IN String DBMUserId,
                                     IN String DBName,
                                     IN String DBUser,
                                     IN String EnvFile,
                                     IN Integer TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sDBHome                   Directory for MaxDB database software.
 $sDBMPassword              Password for the user 'DBMUserId'.
 $sDBMUserId                Database user ID used to connect to MaxDB database server.
 $sDBName                   Name of the MaxDB instance.
 $sDBUser                   Operating system user for MaxDB server.
 $sEnvFile                  Environments file for 'DBUser'. 
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBHome,
       $sDBMPassword,
       $sDBMUserId,
       $sDBName,
       $sDBUser,
       $sEnvFile,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20101, "Subroutine <%s> called with:
  DBHome                   [%s]
  DBMPassword              [%s]
  DBMUserId                [%s]
  DBName                   [%s]
  DBUser                   [%s]
  EnvFile                  [%s]
  TillWhen                 [%s]",

  $sMyName,   $sDBHome,   $sDBMPassword,   $sDBMUserId,   $sDBName,
  $sDBUser, $sEnvFile, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();
  
  #------------------------
  # Return, unless online..
  #------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sCommand = "$sDBHome/dbmcli";
  $sCommandArgs = "-d $sDBName -u $sDBMUserId,$sDBMPassword db_state";

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20102,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( SAPMaxDB::ExecLocalCmd ( $sDBUser, 
                                   $sEnvFile, 
                                   $sCommand, 
                                   $sCommandArgs, 
                                   \$iExitCode, 
                                   $iTimeout, 
                                   \@lsCmdout, 
                                   $sDBMPassword ) ) {
    &logTrace ( 20103,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );

    if ( $lsCmdout [-1] eq 'ONLINE' ) {
      &logTrace ( 20104,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );
    }
    elsif ( $lsCmdout [-1] eq 'ADMIN' ) {
      &logInfo ( 20105,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    }
    elsif ( $lsCmdout [-1] eq 'STARTING' ) {
      &logInfo ( 20106,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    }
    elsif ( $lsCmdout [-1] eq 'OFFLINE' ) {
      &logError ( 20107,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    elsif ( $lsCmdout [-1] eq 'SHUTDOWN' ) {
      &logError ( 20108,
      "<%s> Found search pattern [%s] in the output [%s].",
      $sMyName, $lsCmdout [-1], join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::OFFLINE;
    }
    else {
      &logError ( 20109,
      "<%s> SecondLevelMonitor returned undefined output [%s].",
      $sMyName, join ( chr 012, @lsCmdout ) );

      $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
    }
  }
  elsif ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
    if( $iReturnCode == VCSDef_MonitorExit::TIMEOUT ) {
      &logError ( 20110,
      "<%s> SecondLevelMonitor timedout.", $sMyName );
    }
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String DBUser,
                                  IN String EnvFile,
                                  IN String Cmd,
                                  IN String CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sDBUser                   Operating system user for MaxDB server.
 $sEnvFile                  Environments file for 'DBUser'. 
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
 -------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20111, "Subroutine <%s> called with:
  DBUser                   [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",

  $sMyName, $sDBUser, $sEnvFile, $sCmd, $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20112,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
    &logError ( 20113,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile !~ /\S/ || ! -e $sEnvFile ) {
    &logError ( 20114,
    "<%s> EnvFile [%s] does not exist", $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sDBUser !~ /\S/ || ! defined getpwnam $sDBUser ) {
    &logError ( 20115, "<%s> [%s] is not a valid UNIX login",
    $sMyName, $sDBUser );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
    &logError ( 20116,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20117,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  SAPMaxDB::ExecLocalCmd ( $sDBUser, 
                           $sEnvFile, 
                           $sCmd, 
                           $sCmdArgs, 
                           \$iExitCode, 
                           $iTimeout, 
                           \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
    if ( $iReturnCode == VCSDef_MonitorExit::TIMEOUT ) {
      &logError ( 20118,
      "<%s> Command [%s] with args [%s] timed out", $sMyName, $sCmd, $sCmdArgs );
    }
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20119,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20120,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  if ( $iReturnCode != VCSDef_MonitorExit::TIMEOUT ) {
    $iReturnCode = $hiExitCodes { $iExitCode } || VCSDef_MonitorExit::UNKNOWN;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String PidFile,
                                 In Integer Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20121, "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",

  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20122, "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20123,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20124,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20125,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String PidFile,
                             IN String Proc,
                             IN Integer Uid,
                             OUT RefInt Pid, 
                             OUT RefInt PPid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $iUid           UNIX UID of owner
 $riPid          Reference to existing PID
 $riPPid         Reference to existing PIS's PPID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100          VCSDef_MonitorExit::ONLINE
 -------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $iUid,
       $riPid,
       $riPPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20126, "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  UID                    [%s]
  riPid                  [%s]
  riPPid                 [%s]",

  $sMyName, $sPidFile, $sProc, $iUid, $riPid, $riPPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20127, "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20128, "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20129, "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20130,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20131,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20132,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( SAPMaxDB::GetPidDetails ( $iPid,
                                   \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20133,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( ( $$rhlProc { UID } == $iUid ) &&       # Owner match..
               $$rhlProc { COMMAND } =~ /$sProc\b/ ) { # Filter match.. by vchappav
          &logTrace ( 20134,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
          $$riPPid = $$rhlProc { PPID }; # Store..  # by vchappav ...
        }
        else {
          &logTrace ( 20135,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20136,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20137, "<%s> PID [%s] does not exist",
    $sMyName, $iPid );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool KillSAPMaxDBProcesses ( IN String DBName,
                                         IN String DBUser );

 Description  : Kills processes for this instance of SAPMaxDB
 
 Parameters :

 $sDBName                 Name of the MaxDB instance.
 $sDBUser                 Operating system user for MaxDB server.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub KillSAPMaxDBProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBName,
       $sDBUser ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20138, "Subroutine <%s> called with:
  DBName                   [%s]
  DBUser                   [%s]",

  $sMyName, $sDBName, $sDBUser );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'SAPMaxDB::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    SAPMaxDB::GetProcState ( \@liGoners,
                             $sDBName,
                             $sDBUser,
                             0 ); # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  #@liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20139,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20140,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20141,
    "<%s> Could not locate any process for this instance of SAPMaxDB",
    $sMyName );
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String User,
                                IN String EnvFile,
                                IN String Command,
                                IN String CommandArgs,
                                OUT RefInt ExitCode,
                                IN Integer Timeout,
                                OUT ListReference CmdOut ,
                                IN String HideString );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 $sHideString      String in CommandArgs to hide

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout ,
       $sHideString ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20142, "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",

  $sMyName,
  $sUser, $sEnvFile, $sCommand, $sCommandArgs, $riExitCode,
  $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;

  my ( $iExitCode, $sDispStr, $sCmdDeCr, $sDeCrPasswd ) = ();

  #---------------------
  # Set up for display..
  #---------------------
  $sDispStr = $sCommandArgs;

  #-------------------------
  # Get decrypted password..
  #-------------------------
  if ( defined $sHideString ) {
    #-----------
    # Validate..
    #-----------
    if ( ! -x $sCmdVcsDecrypt ) {
      &logError ( 20143,
      "<%s> Command [%s] is not executable", $sMyName, $sCmdVcsDecrypt );
      return 0; # Failure..
    }

    $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
    ( $sDeCrPasswd ) = grep {
                         s/([^\w])/\\$1/g || 1
                       } readpipe ( "$sCmdDeCr $sHideString" );

    #--------------------
    # Collect exit code..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20144, "<%s> Failed to decrypt crypted string [%s]",
      $sMyName, $sHideString );
      return 0; # Failure..
    }

    $sCommandArgs =~ s/$sHideString/$sDeCrPasswd/g;
  }

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20145,
    "<%s> Cmd [%s] does not exist", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20146,
    "<%s> Cmd [%s] is not executable", $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20147,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sDispStr );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20148,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sDispStr );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgs,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           undef,
                                                           undef,
                                                           $sDeCrPasswd );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20149,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20150,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sDispStr, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20151,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sDispStr, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20152,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName:delete $ENV { LOGNAME };

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20153, "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",

  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20154,
    "<%s> PID [%s] is invalid", $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, $sSwitch, $iExitCode,
       %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------
  if("solaris" eq $sSwitch ){

      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

  }

  elsif("hpux" eq $sSwitch ){

      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

  }

  elsif("aix" eq $sSwitch ){

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';

  }


  elsif("linux" eq $sSwitch ){

      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';

  }

  &logTrace ( 20155, "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20156,
    "<%s> ps(1b) command [%s] is not executable", $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang:delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95:delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20157,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String DBName,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sDBName                   Name of the MaxDB instance.
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDBName,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20158, "Subroutine <%s> called with:
  DBName                   [%s]
  SrvrDetails              [%s]",

  $sMyName, $sDBName, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat, $sEntryPointName ) = ();

  #-------------------------------------
  # List of search patterns to monitor..
  #-------------------------------------
  my @lsSrchPatterns = ( "kernel $sDBName", "kernel $sDBName" );
  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  if (( $::gbStopGlobalListeners) && 
      (($sEntryPointName =~ /clean/) || $sEntryPointName =~ /offline/ )) {
    push ( @lsSrchPatterns, "niserver", "vserver -sdbstarter", "sdbgloballistener", "sdbgloballistener" );
  }
  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20159, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20160, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsSrchPatterns ) {
    $sKey = 'srchstr_' . ++ $iPat;
    $sPidFile = "$sPIDDir/pidfile.$sKey";
    @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
 -------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20161, "Subroutine <%s> called with:
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  $sResourceType = "SAPMaxDB" if $sResourceType !~ /\w/; # Ensure..

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 15,   # Seconds..
                        'OnlineTimeout',  25,
                        'OfflineTimeout', 30,
                        'CleanTimeout',   25 );

  #-----------------------------------------
  # Form query string based on entry point..
  #-----------------------------------------
  $sCommand = $sCmdHatype; # Get global..
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
  $sCommandArgs = "-value $sResourceType $sAttribute";

  #----------------------------------
  # Fire command and collect output..
  #----------------------------------
  ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;

  #--------------------
  # Check exit status..
  #--------------------
  if ( $iExitCode = $? >> 8 ) {
    &logTrace ( 20162,
    "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    $iEngineTimeout = $bUseMargin?$hiGraceMargin{$sAttribute}:0;
  }
  elsif ( ! $iEngineTimeout ) { # Empty value..
    &logTrace ( 20163, "<%s> Could not get [%s] from engine",
    $sMyName, $sAttribute );
    $iEngineTimeout = $bUseMargin?$hiGraceMargin{$sAttribute}:0;
  }
  else { # Success..
    &logTrace ( 20164, "<%s> Got [%s] as [%s] sec from engine",
    $sMyName, $sAttribute, $iEngineTimeout );
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                 # Margin requested..
       $iEngineTimeout >= $hiGraceMargin{$sAttribute} ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin{$sAttribute};       # Apply..

    &logTrace ( 20165,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin{$sAttribute}, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}


=pod

 -------------------------------------------------------------------------
 Function : bool RemoveIPCResourcesForSAPMaxDB ( IN String User,
                                                 IN String DBHome,
                                                 IN String DBName);

 Description : Removes the IPC resources for the MaxDB instance.

 Parameters  : 

 $sUser      User name for which IPC resources are removed.
 $sDBHome    Home directory for SAP MaxDB database instance.
 $sDBName    Database name for which IPC needs to be removed.
 

 Returns : boolean value in the set of {0,1} where
           0 Successfully removed IPCs for SAP MaxDB instance.
           1 Failed to remove IPC resources for SAP MaxDB instance.
 -------------------------------------------------------------------------

=cut

sub RemoveIPCResourcesForSAPMaxDB {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sDBHome,
       $sDBName) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20166, "Subroutine <%s> called with:
  User           [%s]
  LCHome         [%s]
  LCName         [%s]",
  $sMyName, $sUser, $sDBHome, $sDBName);

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + SAPMaxDB::GetEngineTimeout ( 1 );

  #-------------------------------------------------------------
  # Try to get independant data path for SAP MaxDB Server..
  #-------------------------------------------------------------
  my $sCommand = "$sDBHome/dbmcli";
  my $sCommandArgs = " dbm_getpath indepdatapath";

  my ( @lsCmdout, $iExitCode, $iTimeout ) = ();
  my ( $sIndepDataPath, @liIPCshm, @liIPCsem, $sDbDir, $sUsDir, @lsDbFiles, @lsUsFiles ) = ();
  my $iReturnCode = 1;
  my $bDbDirNotFound = 0 ;
  my $bUsDirNotFound = 0 ;

  &logTrace ( 20167,
  "<%s> Getting independent data path for MaxDB instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20168,
    "<%s> Insufficient time to get indepdatapath for MaxDB instance. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    SAPMaxDB::ExecLocalCmd ( "root",
                             "",
                             $sCommand,
                             $sCommandArgs,
                             \$iExitCode,
                             $iTimeout,
                             \@lsCmdout ) ) {
    &logTrace ( 20169,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }
  else {
    &logError ( 20170,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]. Cmd output [%s]",
    $sMyName, $sCommand, $sCommandArgs, $iExitCode,
    join ( chr 012, @lsCmdout ) );
    $iReturnCode = 1;
  }

  if( scalar @lsCmdout ) {
    $sIndepDataPath = $lsCmdout[-1];
    chomp $sIndepDataPath;
  }
  else {
    $sIndepDataPath = "/sapdb/data" ;
    &logError(20171,
              "<%s> Failed to find the indepdatapath with the command [%s] with arguments [%s]",
              $sMyName, $sCommand, $sCommandArgs );
    &logInfo (20172,
              "<%s> Defaulting the indepdatapath for MaxDB server [%s] to [%s]",
              $sMyName, $sDBName, $sIndepDataPath );
  }

  $sDbDir = $sIndepDataPath . "/ipc/" . "db\:" . $sDBName;
  $sUsDir = $sIndepDataPath . "/ipc/" . "us\:" . $sDBName;

  if( ReadDirectory( $sDbDir,
                     \@lsDbFiles ) ) {  # Read from db:<LCName> directory ...
    &logTrace(20173,
              "<%s> Read files [%s] from directory [%s]",
              $sMyName, join(' ', @lsDbFiles), $sDbDir );
    if( scalar @lsDbFiles && SAPMaxDB::GetIPCIDsFromFiles ( \@lsDbFiles,
                                                            O_IPCSEM,
                                                            \@liIPCsem ) ) {
      &logTrace(20174,
                "<%s> Read semaphore IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCsem), $sDbDir );
    }
    else {
      &logTrace(20175,
               "<%s> No semaphore IDs to read from the files in direcotory[%s]",
               $sMyName, $sDbDir );
    }

    if( scalar @lsDbFiles && SAPMaxDB::GetIPCIDsFromFiles ( \@lsDbFiles,
                                                            O_IPCSHM,
                                                            \@liIPCshm ) ) {
      &logTrace(20176,
                "<%s> Read Shared memory IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCshm), $sDbDir );
    }
    else {
      &logTrace(20177,
                "<%s> No Shared Memory IDs to read from the files in direcotory[%s]",
                $sMyName, $sDbDir );
    }
  }
  else {
    &logTrace( 20178,
               "<%s> Failed to read directory [%s].",
               $sMyName, $sDbDir );
    $bDbDirNotFound = 1 ;

  }

  if( SAPMaxDB::ReadDirectory( $sUsDir,
                               \@lsUsFiles ) ) {  # Read from us:<LCName> directory ...
    &logTrace(20179,
              "<%s> Read files [%s] from directory [%s]",
              $sMyName, join(' ', @lsUsFiles), $sUsDir );
    if( scalar @lsUsFiles && SAPMaxDB::GetIPCIDsFromFiles ( \@lsUsFiles,
                                                            O_IPCSEM,
                                                            \@liIPCsem ) ) {
      &logTrace(20180,
                "<%s> Read semaphore IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCsem), $sUsDir );
    }
    else {
      &logTrace(20181,
                "<%s> Failed to read to semaphore IDs from the files in direcotory[%s]",
                $sMyName, $sUsDir );
    }

    if( scalar @lsUsFiles && SAPMaxDB::GetIPCIDsFromFiles ( \@lsUsFiles,
                                                            O_IPCSHM,
                                                            \@liIPCshm ) ) {
      &logTrace(20182,
                "<%s> Read Shared memory IDs [%s] from the files in directory [%s]",
                $sMyName, join(' ', @liIPCshm), $sUsDir );
    }
    else {
      &logTrace(20183,
               "<%s> Failed to read to Shared Memory IDs from the files in direcotory[%s]",
               $sMyName, $sUsDir );
    }
  }
  else {
    &logTrace( 20184,
              "<%s> Failed to read directory [%s].",
              $sMyName, $sUsDir );
    $bUsDirNotFound = 1 ;

  }

  if( scalar @liIPCsem ) {  # Check Semaphores for removing ...
    if( $iReturnCode = SAPMaxDB::RemoveIPCResourcesForUser ( $sUser,
                                                             O_IPCSEM,
                                                             \@liIPCsem ) ) {
      &logTrace (20185,
                 "<%s> Successfully removed Semaphores IDs [%s] for User [%s]",
                 $sMyName, join(' ', @liIPCsem), $sUser );
    }
    else {
      &logError (20186,
                 "<%s> Failed to remove Semaphores for User [%s]",
                 $sMyName, $sUser );
      $iReturnCode = 0; # Failure ...
    }
  }

  if( scalar @liIPCshm ) {  # Check Shared Memory for removing ...
    if( $iReturnCode = SAPMaxDB::RemoveIPCResourcesForUser ( $sUser,
                                                             O_IPCSHM,
                                                             \@liIPCshm ) ) {
      &logTrace (20187,
                 "<%s> Successfully removed Shared Memory IDs [%s] for User [%s]",
                 $sMyName, join(' ', @liIPCshm), $sUser );
    }
    else {
      &logError (20188,
                 "<%s> Failed to remove Shared Memory IDs for User [%s]",
                 $sMyName, $sUser );

      $iReturnCode = 0; # Failure ...
    }
  }

  if( $bDbDirNotFound && $bUsDirNotFound ) {
    $sCommand = "/sapdb/$sDBName/db/pgm/dbmshm";
    $sCommandArgs = " DELETE /var/lib/sdb/dbm/ipc $sDBName";
    &logTrace ( 20230,
    "<%s> Deleting shared memory for MaxDB instance with Cmd [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );

    #------------------
    # Request startup..
    #------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20231,
      "<%s> Insufficient time to delete shared memory for MaxDB instance. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );
      $iReturnCode = 0; # Failure..
    }
    elsif ( $iReturnCode =
      SAPMaxDB::ExecLocalCmd ( "root",
      "",
      $sCommand,
      $sCommandArgs,
      \$iExitCode,
      $iTimeout,
      \@lsCmdout ) ) {
      &logTrace ( 20232,
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
    else {
      &logError ( 20233,
      "<%s> Cmd [%s] with args [%s] failed with exit code [%s]. Cmd output [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode,
      join ( chr 012, @lsCmdout ) );
      $iReturnCode = 1;
    }
  
  }

  return $iReturnCode;
}


=pod

 -------------------------------------------------------------------------
 Function : bool ReadDirectory ( IN String Dir,
                                 IN RefListOfString Files) 

 Description : Reads all the files from the specified directory.

 Parameters  :

 $sDir      The directory from ehich the files needs to be read.
 $rlsFiles  The list of files read from the specified directory.

 Returns : boolean value in the set of {0,1} where
           0 Successfully read all the files from directory.
           1 Failed to read the files from directory.
 -------------------------------------------------------------------------

=cut

sub ReadDirectory {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDir,
       $rlsFiles ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20189, "Subroutine <%s> called with:
  Dir            [%s]
  rlsFiles       [%s]",
  $sMyName, $sDir, $rlsFiles );

  if( ! -d $sDir ) {
    &logTrace ( 20190,
    "<%s> Directory [%s] is not valid. Exiting ...",
    $sMyName, $sDir );

    return 0;  # Failure ...
  }

  my ( $sFile, $sPwd, $sCurr ) = ();

  if ( opendir ( DIRENT, $sDir ) ) {
    &logTrace( 20191,
    "<%s> Directory [%s] opened for reading.",
    $sMyName, $sDir );

    @$rlsFiles = readdir ( DIRENT );

    closedir ( DIRENT ) ||
    &logWarn(20192,
    "<%s> Failed to close directory [%s] after reading",
    $sMyName, $sDir );
  }
  else {
    &logWarn(20193,
    "<%s> Failed to open Directory [%s] for reading.",
    $sMyName, $sDir );

    return 0;  # Failure ...
  }

  return 1;
}


=pod

 -------------------------------------------------------------------------
 Function : bool GetIPCIDsFromFiles ( IN RefListOfString Files,
                                      IN Interger IPCFlags,
                                      OUT RefListOfInteger IPCs );

 Description : Returns all the IPC IDs from the IPC files.

 Parameters  :

 $rlsFiles      List of IPC file names from which the ID are to read.
 $riIPCFlags    The type of IPCs needs to be read, like O_IPCSEM, O_IPCSHM etc..
 $rliIPCs       List of IPC IDs read from the files.

 Returns : boolean value in the set of {0,1} where
           0 Successfully read IDs from files.
           1 Failed to read IDs from files.
 -------------------------------------------------------------------------

=cut

sub GetIPCIDsFromFiles {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rlsFiles,
       $riIPCFlags,
       $rliIPCs ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20194, "Subroutine <%s> called with:
  rlsFiles       [%s]
  IPC Flags      [%s]
  rliIPCs        [%s]",
  $sMyName, $rlsFiles, $riIPCFlags, $rliIPCs );

  foreach my $sFile (@$rlsFiles) {
    chomp $sFile;

    my $iIPC = $sFile;

    if( O_IPCSHM == $riIPCFlags && $iIPC =~ s/^-m//o ) {
      &logTrace(20195,
      "<%s> Got Shared Memory ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
    elsif( O_IPCSEM == $riIPCFlags && $iIPC =~ s/^-s//o ) {
      &logTrace(20196,
      "<%s> Got Semaphore ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
    elsif( O_IPCMSQ == $riIPCFlags && $iIPC =~ s/^-q//o ) {
      &logTrace(20197,
      "<%s> Got Message Queue ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
    elsif( O_IPCALL == $riIPCFlags && $iIPC =~ s/^-[sqm]+//o ) {
      &logTrace(20198,
      "<%s> Got IPC ID [%s] from [%s].",
      $sMyName, $iIPC, $sFile );

      push(@$rliIPCs, $iIPC);
    }
  }

  return 1;
}


=pod

 -------------------------------------------------------------------------
 Function : bool RemoveIPCResourcesForUser ( IN String User,
                                             IN Integer IpcFlags,
                                             IN RefListInteger Ids );

 Description : Removes IPC resources for specified UNIX user

 Parameters :

 $sUser      UNIX user whose IPC resources are to be removed
 $iIpcFlags  Behaviour flag
 $rliIds     List of IPC IDs to remove for that user

 Values for Behaviour flag:
 O_IPCSHM : Shared Memory           O_IPCSEM : Semaphore
 O_IPCMSQ : Message Queue           O_IPCALL : All of the above

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
 -------------------------------------------------------------------------

=cut

sub RemoveIPCResourcesForUser {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $iIpcFlags,
       $rliIds ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace( 20199, "Subroutine <%s> called with:
  User           [%s]
  IPC Flags      [%s]
  rliIds         [%s]",
  $sMyName, $sUser, $iIpcFlags, $rliIds );

  #-------------------------
  # Define local variables..
  #-------------------------
  my $sFormat = 'l l l l ul ul i'; # Needed for unpack. Check intro(2)..
  my $iLocn = $^O eq 'linux'?1:0;  # Index locn. of ipc_perm.uid..
  my $lsVar = 0;                   # Stores the binary structure semid_ds..
  my $sCommand = $sCmdIpcs;        # Get global..

  my ( $iId, $iUid, $iIpcUid, $iRemoved,
       @liShmIds, @liMqIds, @liSemIds, @liIdsRemoved, %hiIds ) = ();

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------
  # Validate the user..
  #--------------------
  unless ( defined ( $iUid = getpwnam $sUser ) ) {
    &logError ( 20200, "<%s> Invalid User [%s] specfied. Cannot proceed.",
    $sMyName, $sUser );
    return 0; # Error..
  }
  #----------------
  # Check for IDs..
  #----------------
  if ( scalar @$rliIds ) {
    &logTrace ( 20201,
    "<%s> Request to remove only [%s] IDs for User [%s]",
    $sMyName, join ( ' ', @$rliIds ), $sUser );

    %hiIds = map { $_ => 1 } @$rliIds; # Remove duplicates..
  }
#  else {
#    &logWarn ( 20032,
#    "<%s> Called with [%s] IDs for User [%s]",
#    $sMyName, join ( ' ', @$rliIds ), $sUser );
#    return 1;
#  }

  #-------------------------------
  # Read in the contents of IPCs..
  #-------------------------------
  foreach ( readpipe ( $sCommand ) ) {
    s/\n//; # Chomp..

    #---------------
    # Handle Linux..
    #---------------
    if ( $^O eq 'linux' ) {
      if ( $iIpcFlags & O_IPCSHM && # Shared Memory..
           /shmid/ ... /^$/ ) {
        /(\w+)\s+(\d+)/;
        push ( @liShmIds, $2 ) if defined $2;
      }

      if ( $iIpcFlags & O_IPCSEM && # Semaphore.. 
           /semid/ ... /^$/ ) {
        /(\w+)\s+(\d+)/;
        push ( @liSemIds, $2 ) if defined $2;
      }

      if ( $iIpcFlags & O_IPCMSQ && #  Message Queue..
           /msqid/ ... /^$/ ) {
        /(\w+)\s+(\d+)/;
        push ( @liMqIds, $2 ) if defined $2;
      }
    }
    else { # the rest..
      #--------------------
      # Skip removed IPCs..
      #--------------------
      if ( /^[m|s|q]\s*(\d+)\s+\w+\s+(\S)/ && # Of all three..
           defined $2                      && # Mode defined..
           ( $2 eq 'D' || $2 eq 'C' ) ) {     # Says, it's removed..
        next;                                 # Skip..
      }

      $iIpcFlags & O_IPCSHM && # Shared Memory..
      push ( @liShmIds, $_ =~ /^m\s*(\d+).*/ );

      $iIpcFlags & O_IPCMSQ && # Message Queue..
      push ( @liMqIds,  $_ =~ /^q\s*(\d+).*/ );

      $iIpcFlags & O_IPCSEM && # Semaphore..
      push ( @liSemIds, $_ =~ /^s\s*(\d+).*/ );
    }
  }

  #---------------------------------------------
  # Process each one separately as per request..
  #---------------------------------------------

  #----------------
  # Shared Memory..
  #----------------
  if ( $iIpcFlags & O_IPCSHM ) {
    foreach $iId ( @liShmIds ) {

      next if ( scalar @$rliIds && ! $hiIds { $iId } );

      shmctl( $iId, IPC_STAT, $lsVar ) ||
      &logError ( 20202,
      "<%s> Failed to use shmctl(2) on SHM Id [%s]", $sMyName, $iId );

      #-----------------------------------------------------
      # Unpack uid(2) out of the packed semid_ds structure..
      #-----------------------------------------------------
      $iIpcUid = ( unpack ( $sFormat, $lsVar ) ) [$iLocn];

      #-----------------------------
      # Remove resource if a match..
      #-----------------------------
      if ( $iIpcUid == $iUid ) {
        if ( shmctl ( $iId, IPC_RMID, 0 ) ) {
          push ( @liIdsRemoved, $iId );
          ++$iRemoved; # Increment..
        }
        else {
          &logError ( 20203,
          "<%s> Failed to remove SHM Id [%s]. Error [%s]", $sMyName, $iId, $! );
          $iReturnCode = 0; # Failed..
        }
      }
    }
  }
  if ( scalar @liIdsRemoved ) {
    &logTrace ( 20204, "<%s>
    +--- Removed SHM Ids: [%s] for User [%s] Uid [%s]. Total [%s] ---+",
    $sMyName, join ( ' ', @liIdsRemoved ), $sUser, $iUid, scalar @liIdsRemoved );
    @liIdsRemoved = (); # Clear..
  }

  #----------------
  # Message Queue..
  #----------------
  $lsVar = 0; # Reset..
  if ( $iIpcFlags & O_IPCMSQ ) {
    foreach $iId ( @liMqIds ) {

      next if ( scalar @$rliIds && ! $hiIds { $iId } );

      msgctl( $iId, IPC_STAT, $lsVar ) ||
      &logError ( 20205,
      "<%s> Failed to use msgctl(2) on SHM Id [%s]", $sMyName, $iId );

      #-----------------------------------------------------
      # Unpack uid(2) out of the packed semid_ds structure..
      #-----------------------------------------------------
      $iIpcUid = ( unpack ( $sFormat, $lsVar ) ) [$iLocn];

      #-----------------------------
      # Remove resource if a match..
      #-----------------------------
      if ( $iIpcUid == $iUid ) {
        if ( msgctl ( $iId, IPC_RMID, 0 ) ) {
          push ( @liIdsRemoved, $iId );
          ++$iRemoved; # Increment..
        }
        else {
          &logError ( 20206,
          "<%s> Failed to remove Msg QId [%s]. Error [%s]", $sMyName, $iId, $! );
          $iReturnCode = 0; # Failed..
        }
      }
    }
  }

  if ( scalar @liIdsRemoved ) {
    &logTrace ( 20207, "<%s>
    +--- Removed Msg QIds: [%s] for User [%s] Uid [%s]. Total [%s] ---+",
    $sMyName, join ( ' ', @liIdsRemoved ), $sUser, $iUid, scalar @liIdsRemoved );
    @liIdsRemoved = (); # Clear..
  }

  #------------
  # Semaphore..
  #------------
  $lsVar = 0; # Reset..
  if ( $iIpcFlags & O_IPCSEM ) {
    foreach $iId ( @liSemIds ) {

      next if ( scalar @$rliIds && ! $hiIds { $iId } );

      semctl( $iId, 0, IPC_STAT, $lsVar ) ||
      &logError ( 20208,
      "<%s> Failed to use semctl(2) on Sem Id [%s]", $sMyName, $iId );

      #-----------------------------------------------------
      # Unpack uid(2) out of the packed semid_ds structure..
      #-----------------------------------------------------
      $iIpcUid = ( unpack ( $sFormat, $lsVar ) ) [$iLocn];

      #-----------------------------
      # Remove resource if a match..
      #-----------------------------
      if ( $iIpcUid == $iUid ) {
        if ( semctl ( $iId, 0, IPC_RMID, 0 ) ) {
          push ( @liIdsRemoved, $iId );
          ++$iRemoved; # Increment..
        }
        else {
          &logError ( 20209,
          "<%s> Failed to remove Sem Id [%s]. Error [%s]", $sMyName, $iId, $! );
          $iReturnCode = 0; # Failed..
        }
      }
    }
  }

  if ( scalar @liIdsRemoved ) {
    &logTrace ( 20210,
    "<%s>
    +--- Removed Sem Ids: [%s] for User [%s] Uid [%s]. Total [%s] ---+",
    $sMyName, join ( ' ', @liIdsRemoved ), $sUser, $iUid, scalar @liIdsRemoved );
    @liIdsRemoved = (); # Clear..
  }

  #---------------------------------
  # Log, if we did anything useful..
  #---------------------------------
  if ( $iRemoved ) {
    &logTrace( 20211,
    "<%s> Successfully removed [%s] IPC resources for User [%s] in all.",
    $sMyName, $iRemoved, $sUser );
  }

  return $iReturnCode;
}


1; # So PERL knows its included..
